@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019-2021 Job Snijders <job@sobornost.net>
+ * Copyright (c) 2019-2022 Job Snijders <job@sobornost.net>
  * Copyright (c) 2018 Peter Schoenmaker <pds@ntt.net>
  * Copyright (c) 2007-2019 Alexandre Snarskii <snar@snar.spb.ru>
  * All rights reserved.
@@ -86,6 +86,7 @@ bgpq_expander_init(struct bgpq_expander *b, int af)
 
 	b->family = af;
 	b->sources = "";
+	b->usesource = 0;
 	b->name = "NN";
 	b->aswidth = 8;
 	b->identify = 1;
@@ -152,7 +153,7 @@ bgpq_expander_add_already(struct bgpq_expander *b, char *rs)
 	lkey.text = rs;
 
 	if (RB_FIND(tentree, &b->already, &lkey))
-		return 1;
+		return 0;
 
 	le = sx_tentry_new(rs);
 
@@ -169,7 +170,7 @@ bgpq_expander_add_stop(struct bgpq_expander *b, char *rs)
 	lkey.text = rs;
 
 	if (RB_FIND(tentree, &b->stoplist, &lkey))
-		return 1;
+		return 0;
 
 	le = sx_tentry_new(rs);
 
@@ -196,8 +197,10 @@ bgpq_expander_add_as(struct bgpq_expander *b, char *as)
 	}
 
 	if (!expand_special_asn &&
-	    ((asno  >= 4200000000ul) || (asno >= 64496 && asno <= 65551)))
+	    (asno >= 4200000000ul || (asno >= 64496 && asno <= 65551))) {
+		sx_report(SX_ERROR,"Invalid AS number: %u\n", asno);
 		return 0;
+	}
 
 	if ((asne = malloc(sizeof(struct asn_entry))) == NULL)
 		err(1, NULL);
@@ -229,7 +232,7 @@ bgpq_expander_add_prefix(struct bgpq_expander *b, char *prefix)
 	sx_radix_tree_insert(b->tree, p);
 
         if (p)
-		sx_prefix_destroy(p);
+		sx_prefix_free(p);
 
 	return 1;
 }
@@ -240,6 +243,63 @@ bgpq_expander_add_prefix_range(struct bgpq_expander *b, char *prefix)
 	return sx_prefix_range_parse(b->tree, b->family, b->maxlen, prefix);
 }
 
+char *
+bgpq_get_asset(char *object) {
+	char *d, *asset;
+
+	d = strstr(object, "::");
+	if (d)
+		d += 2;
+	else
+		d = object;
+
+	if ((asset = calloc(1, 256)) == NULL)
+		err(1, NULL);
+
+	memcpy(asset, d, strlen(object) - (d - object));
+
+	return asset;
+}
+
+char *
+bgpq_get_rset(char *object) {
+	char *d, *rset;
+
+	d = strstr(object, "::");
+
+	if (d)
+		d += 2;
+	else
+		d = object;
+
+	if ((rset = calloc(1, 256)) == NULL)
+		err(1, NULL);
+
+	memcpy(rset, d, strlen(object) - (d - object));
+
+	return rset;
+}
+
+char *
+bgpq_get_source(char *object) {
+	char 		*d;
+	char 		*source;
+	unsigned int	 slen;
+
+	d = strstr(object, "::");
+
+	if (d) {
+		if ((source = calloc(1, 256)) == NULL)
+			err(1, NULL);
+
+		slen = d - object;
+		memcpy(source, object, slen);
+		return source;
+	}
+
+	return NULL;
+}
+
 static int
 bgpq_expanded_macro(char *as, struct bgpq_expander *ex,
     struct request *req)
@@ -249,11 +309,13 @@ bgpq_expanded_macro(char *as, struct bgpq_expander *ex,
 	return 1;
 }
 
-struct request *bgpq_pipeline(struct bgpq_expander *b,
+struct request *
+bgpq_pipeline(struct bgpq_expander *b,
     int (*callback)(char *, struct bgpq_expander *b, struct request *req),
     void *udata, char *fmt, ...);
 
-int bgpq_expand_irrd(struct bgpq_expander *b,
+int
+bgpq_expand_irrd(struct bgpq_expander *b,
     int (*callback)(char*, struct bgpq_expander *b, struct request *req),
     void *udata, char *fmt, ...);
 
@@ -261,6 +323,9 @@ static int
 bgpq_expanded_macro_limit(char *as, struct bgpq_expander *b,
     struct request *req)
 {
+	char		*source;
+	struct request	*req1;
+
 	if (!strncasecmp(as, "AS-", 3) || strchr(as, '-') || strchr(as, ':')) {
 		struct sx_tentry tkey = { .text = as };
 
@@ -281,20 +346,41 @@ bgpq_expanded_macro_limit(char *as, struct bgpq_expander *b,
 		    req->depth + 1 < b->maxdepth)) {
 			bgpq_expander_add_already(b, as);
 			if (pipelining) {
-				struct request *req1 = bgpq_pipeline(b,
-				    bgpq_expanded_macro_limit, NULL, "!i%s\n",
-				        as);
+				if (b->usesource) {
+					source = bgpq_get_source(as);
+					if (source) {
+						bgpq_pipeline(b, NULL, NULL,
+						    "!s%s\n", source);
+						free(source);
+					} else {
+						bgpq_pipeline(b, NULL, NULL,
+						    "!s%s\n", b->defaultsources);
+					}
+				}
+
+				req1 = bgpq_pipeline(b, bgpq_expanded_macro_limit,
+				    NULL, "!i%s\n", bgpq_get_asset(as));
 				req1->depth = req->depth + 1;
 			} else {
+				if (b->usesource) {
+					source = bgpq_get_source(as);
+					if (source) {
+						bgpq_expand_irrd(b, NULL, NULL,
+						    "!s%s\n", source);
+						free(source);
+					} else {
+						bgpq_expand_irrd(b, NULL, NULL,
+						    "!s%s\n", b->defaultsources);
+					}
+				}
 				b->cdepth++;
 				bgpq_expand_irrd(b, bgpq_expanded_macro_limit,
-				    NULL, "!i%s\n", as);
+				    NULL, "!i%s\n", bgpq_get_asset(as));
 				b->cdepth--;
 			}
 		} else {
 			SX_DEBUG(debug_expander > 2, "ignoring %s at depth %i\n",
-			    as,
-			    b->cdepth ? (b->cdepth + 1) : (req->depth + 1));
+			    as, b->cdepth ? (b->cdepth + 1) : (req->depth + 1));
 		}
 	} else if (!strncasecmp(as, "AS", 2)) {
 		struct sx_tentry tkey = { .text = as };
@@ -350,13 +436,88 @@ bgpq_expanded_v6prefix(char *prefix, struct bgpq_expander *ex,
 	return 1;
 }
 
-int bgpq_pipeline_dequeue(int fd, struct bgpq_expander *b);
+static char *
+bgpq_get_irrd_sources(int fd)
+{
+	char			*query, *response, *sources, *start, *end;
+	const unsigned int	 rsize = 256;
+	unsigned int		 slen;
+	int			 ret, qlen;
+
+	query = "!s-lc\n";
+	qlen = strlen(query);
+
+	if ((response = calloc(1, rsize)) == NULL)
+		err(1, NULL);
+	if ((sources = calloc(1, rsize)) == NULL)
+		err(1, NULL);
+
+	SX_DEBUG(debug_expander, "Requesting source list %s", query);
+	if ((ret = write(fd, query, strlen(query))) != qlen) {
+		sx_report(SX_ERROR, "Partial write of query to "
+			"IRRd: %i bytes, %s\n", ret, strerror(errno));
+		close(fd);
+		free(sources);
+		free(response);
+		exit(1);
+	}
+
+	if (0 < read(fd, response, rsize)) {
+		SX_DEBUG(debug_expander, "Got answer %s", response);
+		if (*(response + strlen(response) - 2) != 'C') {
+			sx_report(SX_ERROR, "Invalid response "
+				"'%s': %s\n", response, query);
+			close(fd);
+			free(sources);
+			free(response);
+			exit(1);
+		}
+	} else {
+		sx_report(SX_ERROR, "failed to read sources\n");
+		close(fd);
+		free(sources);
+		free(response);
+		exit(1);
+	}
+
+	start = strchr(response, '\n');
+	if (start) {
+		start += 1;
+		end = strchr(start, '\n');
+		if (!end) {
+			sx_report(SX_ERROR, "No 2nd newline in response '%s': %s\n",
+				response, query);
+			close(fd);
+			free(sources);
+			free(response);
+			exit(1);
+		}
+		slen = end - start;
+		if (slen > rsize)
+			memcpy(sources, start, rsize - 1);
+		else
+			memcpy(sources, start, slen);
+	} else {
+		sx_report(SX_ERROR, "No 1st newline in response '%s': %s\n",
+			response, query);
+		close(fd);
+		free(sources);
+		free(response);
+		exit(1);
+	}
+
+	free(response);
+	return sources;
+}
 
 static struct request *
 request_alloc(char *request, int (*callback)(char *, struct bgpq_expander *,
     struct request *), void *udata)
 {
-	struct request *bp = malloc(sizeof(struct request));
+	struct request *bp;
+
+	if ((bp = malloc(sizeof(struct request))) == NULL)
+		err(1, NULL);
 
 	if (!bp)
 		return NULL;
@@ -385,9 +546,9 @@ bgpq_pipeline(struct bgpq_expander *b,
     int (*callback)(char *, struct bgpq_expander *, struct request *),
     void *udata, char *fmt, ...)
 {
-	char			 request[128];
+	struct request		*bp = NULL;
+	char			 request[256];
 	int			 ret;
-	struct request	*bp = NULL;
 	va_list			 ap;
 
 	va_start(ap, fmt);
@@ -402,7 +563,6 @@ bgpq_pipeline(struct bgpq_expander *b,
 		sx_report(SX_FATAL,"Unable to allocate %lu bytes: %s\n",
 		    (unsigned long)sizeof(struct request),
 		    strerror(errno));
-		exit(1);
 	}
 
 	if (STAILQ_EMPTY(&b->wq)) {
@@ -415,12 +575,14 @@ bgpq_pipeline(struct bgpq_expander *b,
 			sx_report(SX_FATAL, "Error writing request: %s\n",
 			    strerror(errno));
 		}
+
 		bp->offset=ret;
-		if (ret == bp->size) {
+
+		if (ret == bp->size)
 			STAILQ_INSERT_TAIL(&b->rq, bp, next);
-		} else {
+		else
 			STAILQ_INSERT_TAIL(&b->wq, bp, next);
-		}
+
 	} else
 		STAILQ_INSERT_TAIL(&b->wq, bp, next);
 
@@ -528,6 +690,7 @@ bgpq_read(struct bgpq_expander *b)
 {
 	static char response[256];
 	static int off = 0;
+	int rval = 1;
 
 	if (!STAILQ_EMPTY(&b->wq))
 		bgpq_write(b);
@@ -568,10 +731,8 @@ have:
 			unsigned long 	 togot = strtoul(response + 1, &eon, 10);
 			char 		*recvbuffer = malloc(togot + 2);
 
-			if (!recvbuffer) {
-				sx_report(SX_FATAL, "error allocating %lu "
-				    "bytes: %s\n", togot + 2, strerror(errno));
-			}
+			if (recvbuffer == NULL)
+				err(1, NULL);
 
 			memset(recvbuffer, 0, togot + 2);
 
@@ -666,7 +827,7 @@ have3:
 					c[spn] = 0;
 				if (c[0] == 0)
 					break;
-				req->callback(c, b, req);
+				if (!req->callback(c, b, req)) rval = 0;
 				c += spn + 1;
 			}
 			assert(c == recvbuffer + togot);
@@ -674,24 +835,26 @@ have3:
 			free(recvbuffer);
 		} else if (response[0] == 'C') {
 			/* No data */
-			SX_DEBUG(debug_expander,"No data expanding %s\n",
+			SX_DEBUG(debug_expander,"No data expanding %s",
 			    req->request);
 			if (b->validate_asns)
 				bgpq_expander_invalidate_asn(b, req->request);
 		} else if (response[0] == 'D') {
-			/* .... */
-			SX_DEBUG(debug_expander,"Key not found expanding %s\n",
+			SX_DEBUG(debug_expander, "Key not found expanding %s",
 			    req->request);
 			if (b->validate_asns)
 				bgpq_expander_invalidate_asn(b, req->request);
+			rval = 0;
 		} else if (response[0] == 'E') {
-			sx_report(SX_ERROR, "Multiple keys expanding %s: %s\n",
+			sx_report(SX_ERROR, "Multiple keys expanding %s: %s",
 			    req->request, response);
+			rval = 0;
 		} else if ( response[0] == 'F') {
-			sx_report(SX_ERROR, "Error expanding %s: %s\n",
+			sx_report(SX_ERROR, "Error expanding %s: %s",
 			    req->request, response);
+			rval = 0;
 		} else {
-			sx_report(SX_ERROR,"Wrong reply: %s to %s\n", response,
+			sx_report(SX_ERROR,"Wrong reply: %s to %s", response,
 			    req->request);
 			exit(1);
 		}
@@ -708,7 +871,7 @@ have3:
 		request_free(req);
 	}
 
-	return 0;
+	return rval;
 }
 
 int
@@ -716,11 +879,12 @@ bgpq_expand_irrd(struct bgpq_expander *b,
     int (*callback)(char *, struct bgpq_expander *, struct request *),
     void *udata, char *fmt, ...)
 {
-	char			 request[128], response[128];
+	char			 request[256], response[256];
 	va_list			 ap;
 	ssize_t			 ret;
 	int			 off = 0;
 	struct request	*req;
+	int rval = 1;
 
 	va_start(ap, fmt);
 	vsnprintf(request, sizeof(request), fmt, ap);
@@ -730,8 +894,12 @@ bgpq_expand_irrd(struct bgpq_expander *b,
 
 	SX_DEBUG(debug_expander, "expander sending: %s", request);
 
-	if ((ret = write(b->fd, request, strlen(request)) == 0) || ret == -1)
-		err(1, "write");
+	if ((ret = write(b->fd, request, strlen(request)) == 0) || ret == -1) {
+		sx_report(SX_ERROR,
+			"Partial write of request to IRRd: %li bytes, %s\n",
+			ret, strerror(errno));
+		exit(1);
+	}
 
 	memset(response, 0, sizeof(response));
 
@@ -743,7 +911,6 @@ repeat:
 		exit(1);
 	} else if (ret == 0) {
 		sx_report(SX_FATAL, "EOF reading IRRd\n");
-		exit(1);
 	}
 
 	off += ret;
@@ -808,10 +975,8 @@ reread2:
 		if (ret < 0) {
 			sx_report(SX_FATAL, "error reading IRRd: %s\n",
 			    strerror(errno));
-			exit(1);
 		} else if (ret == 0) {
 			sx_report(SX_FATAL, "eof reading IRRd\n");
-			exit(1);
 		}
 		off += ret;
 
@@ -831,7 +996,7 @@ have3:
 			if (c[0] == 0)
 				break;
 			if (callback)
-				callback(c, b, req);
+				if (!callback(c, b, req)) rval = 0;
 			c += spn + 1;
 		}
 		memset(recvbuffer, 0, togot + 2);
@@ -841,30 +1006,36 @@ have3:
 		if (b->validate_asns)
 			bgpq_expander_invalidate_asn(b, request);
 	} else if (response[0] == 'D') {
-		/* ... */
+		SX_DEBUG(debug_expander, "Key not found expanding %s",
+			req->request);
 		if (b->validate_asns)
 			bgpq_expander_invalidate_asn(b, request);
+		rval = 0;
 	} else if (response[0] == 'E') {
 		/* XXXXXX */
 	} else if (response[0] == 'F') {
-		/* XXXXXX */
+		sx_report(SX_ERROR, "Error expanding %s: %s",
+			request, response);
+		rval = 0;
 	} else {
-		sx_report(SX_ERROR,"Wrong reply: %s\n", response);
-		exit(0);
+		sx_report(SX_ERROR,"Wrong reply: %s", response);
+		exit(1);
 	}
 	request_free(req);
 
-	return 0;
+	return rval;
 }
 
 int
 bgpq_expand(struct bgpq_expander *b)
 {
-	int			 fd = -1, err, ret, aquery = 0;
+	char			*source;
 	struct slentry		*mc;
 	struct addrinfo 	 hints, *res = NULL, *rp;
 	struct linger		 sl;
 	struct asn_entry	*asne;
+	int			 fd = -1, err, ret, aquery = 0;
+	int			 slen;
 
 	sl.l_onoff = 1;
 	sl.l_linger = 5;
@@ -931,6 +1102,7 @@ bgpq_expand(struct bgpq_expander *b)
 	if ((ret = write(fd, "!!\n", 3)) != 3) {
 		sx_report(SX_ERROR, "Partial write of multiple command mode "
 		    "to IRRd: %i bytes, %s\n", ret, strerror(errno));
+		close(fd);
 		exit(1);
 	}
 
@@ -944,6 +1116,7 @@ bgpq_expand(struct bgpq_expander *b)
 				sx_report(SX_ERROR, "Partial write of "
 				    "identifier to IRRd: %i bytes, %s\n",
 				    ret, strerror(errno));
+				close(fd);
 				exit(1);
 			}
 			memset(ident, 0, sizeof(ident));
@@ -951,10 +1124,12 @@ bgpq_expand(struct bgpq_expander *b)
 				SX_DEBUG(debug_expander, "Got answer %s", ident);
 			} else {
 				sx_report(SX_ERROR, "ident, failed read from IRRd\n");
+				close(fd);
 				exit(1);
 			}
 		} else {
 			sx_report(SX_ERROR, "snprintf(ident) failed\n");
+			close(fd);
 			exit(1);
 		}
 	}
@@ -967,6 +1142,7 @@ bgpq_expand(struct bgpq_expander *b)
 		if ((ret = write(fd, "!a\n", 3)) != 3) {
 			sx_report(SX_ERROR, "Partial write of '!a' test query "
 			    "to IRRd: %i bytes, %s\n", ret, strerror(errno));
+			close(fd);
 			exit(1);
 		}
 		memset(aret, 0, sizeof(aret));
@@ -979,14 +1155,27 @@ bgpq_expand(struct bgpq_expander *b)
 			}
 		} else {
 			sx_report(SX_ERROR, "A query test failed read from IRRd\n");
+			close(fd);
 			exit(1);
 		}
 	}
 
+	if (b->usesource) {
+		if (b->sources && b->sources[0] != 0) {
+			b->defaultsources = (char*)calloc(1, strlen(b->sources));
+			strlcpy(b->defaultsources, b->sources,
+			    sizeof(b->defaultsources));
+		} else {
+			b->defaultsources = bgpq_get_irrd_sources(b->fd);
+		}
+	} else {
+		b->defaultsources = bgpq_get_irrd_sources(b->fd);
+	}
+
 	if (b->sources && b->sources[0] != 0) {
-		int slen = strlen(b->sources) + 4;
-		if (slen < 128)
-			slen = 128;
+		slen = strlen(b->sources) + 4;
+		if (slen < 256)
+			slen = 256;
 		char sources[slen];
 		slen = snprintf(sources, sizeof(sources), "!s%s\n", b->sources);
 		if (slen > 0) {
@@ -994,6 +1183,7 @@ bgpq_expand(struct bgpq_expander *b)
 			if ((ret = write(fd, sources, slen)) != slen) {
 				sx_report(SX_ERROR, "Partial write of sources to "
 				    "IRRd: %i bytes, %s\n", ret, strerror(errno));
+				close(fd);
 				exit(1);
 			}
 			memset(sources, 0, sizeof(sources));
@@ -1002,14 +1192,17 @@ bgpq_expand(struct bgpq_expander *b)
 				if (sources[0] != 'C') {
 					sx_report(SX_ERROR, "Invalid source(s) "
 					    "'%s': %s\n", b->sources, sources);
+					close(fd);
 					exit(1);
 				}
 			} else {
 				sx_report(SX_ERROR, "failed to read sources\n");
+				close(fd);
 				exit(1);
 			}
 		} else {
 			sx_report(SX_ERROR, "snprintf(sources) failed\n");
+			close(fd);
 			exit(1);
 		}
 	}
@@ -1019,25 +1212,57 @@ bgpq_expand(struct bgpq_expander *b)
 
 	STAILQ_FOREACH(mc, &b->macroses, entry) {
 		if (!b->maxdepth && RB_EMPTY(&b->stoplist)) {
-			if (aquery)
+			if (b->usesource) {
+				source = bgpq_get_source(mc->text);
+				if (source){
+					if (pipelining){
+						bgpq_pipeline(b, NULL, NULL, "!s%s\n", source);
+						bgpq_pipeline(b, bgpq_expanded_macro_limit, b,
+							"!i%s\n", bgpq_get_asset(mc->text));
+					} else {
+						bgpq_expand_irrd(b, NULL, NULL, "!s%s\n", source);
+						bgpq_expand_irrd(b, bgpq_expanded_macro_limit,
+							b, "!i%s\n", bgpq_get_asset(mc->text));
+					}
+					free(source);
+				} else {
+					if (pipelining){
+						bgpq_pipeline(b, NULL, NULL, "!s%s\n",
+							b->defaultsources);
+						bgpq_pipeline(b, bgpq_expanded_macro_limit, b,
+							"!i%s\n", bgpq_get_asset(mc->text));
+					} else {
+						bgpq_expand_irrd(b, NULL, NULL, "!s%s\n",
+							b->defaultsources);
+						bgpq_expand_irrd(b, bgpq_expanded_macro_limit, b,
+							"!i%s\n", bgpq_get_asset(mc->text));
+					}
+				}
+			} else if (aquery)
 				bgpq_expand_irrd(b, bgpq_expanded_prefix, b,
 				    "!a%s%s\n",
 				    b->family == AF_INET ? "4" : "6",
-				    mc->text);
+				    bgpq_get_asset(mc->text));
 			else
 				bgpq_expand_irrd(b, bgpq_expanded_macro, b,
-				    "!i%s,1\n", mc->text);
+				    "!i%s,1\n", bgpq_get_asset(mc->text));
 		} else {
-			bgpq_expander_add_already(b, mc->text);
+			bgpq_expander_add_already(b, bgpq_get_asset(mc->text));
 			if (pipelining)
 				bgpq_pipeline(b, bgpq_expanded_macro_limit,
-				    NULL, "!i%s\n", mc->text);
+				    NULL, "!i%s\n", bgpq_get_asset(mc->text));
 			else
 				bgpq_expand_irrd(b, bgpq_expanded_macro_limit,
-				    NULL, "!i%s\n", mc->text);
+				    NULL, "!i%s\n", bgpq_get_asset(mc->text));
 		}
 	}
 
+	if (pipelining){
+		bgpq_pipeline(b, NULL, NULL, "!s%s\n", b->defaultsources);
+	} else {
+		bgpq_expand_irrd(b, NULL, NULL, "!s%s\n", b->defaultsources);
+	}
+
 	if (pipelining) {
 		if (!STAILQ_EMPTY(&b->wq))
 			bgpq_write(b);
@@ -1047,12 +1272,70 @@ bgpq_expand(struct bgpq_expander *b)
 
 	if (b->generation >= T_PREFIXLIST || b->validate_asns) {
 		STAILQ_FOREACH(mc, &b->rsets, entry) {
-			if (b->family == AF_INET)
-				bgpq_expand_irrd(b, bgpq_expanded_prefix,
-				    NULL, "!i%s,1\n", mc->text);
-			else
-				bgpq_expand_irrd(b, bgpq_expanded_v6prefix,
-				    NULL, "!i%s,1\n", mc->text);
+			if (b->usesource) {
+				source = bgpq_get_source(mc->text);
+				if (source){
+					if (pipelining){
+						printf("Checking %s\n", bgpq_get_rset(mc->text));
+						bgpq_pipeline(b, NULL, NULL, "!s%s\n", source);
+						if (b->family == AF_INET)
+							bgpq_pipeline(b, bgpq_expanded_prefix,
+				    			NULL, "!i%s\n", bgpq_get_rset(mc->text));
+						else
+							bgpq_pipeline(b, bgpq_expanded_v6prefix,
+								NULL, "!i%s\n", bgpq_get_rset(mc->text));
+					} else {
+						bgpq_expand_irrd(b, NULL, NULL, "!s%s\n", source);
+						if (b->family == AF_INET)
+							bgpq_expand_irrd(b, bgpq_expanded_prefix,
+				    			NULL, "!i%s\n", bgpq_get_rset(mc->text));
+						else
+							bgpq_expand_irrd(b, bgpq_expanded_v6prefix,
+								NULL, "!i%s\n", bgpq_get_rset(mc->text));
+					}
+					free(source);
+				} else {
+					if (pipelining){
+						bgpq_pipeline(b, NULL, NULL, "!s%s\n",
+							b->defaultsources);
+						if (b->family == AF_INET)
+							bgpq_pipeline(b, bgpq_expanded_prefix,
+								NULL, "!i%s\n", bgpq_get_rset(mc->text));
+						else
+							bgpq_pipeline(b, bgpq_expanded_v6prefix,
+								NULL, "!i%s\n", bgpq_get_rset(mc->text));
+					} else {
+						bgpq_expand_irrd(b, NULL, NULL, "!s%s\n",
+							b->defaultsources);
+						if (b->family == AF_INET)
+							bgpq_expand_irrd(b, bgpq_expanded_prefix,
+								NULL, "!i%s\n", bgpq_get_rset(mc->text));
+						else
+							bgpq_expand_irrd(b, bgpq_expanded_v6prefix,
+								NULL, "!i%s\n", bgpq_get_rset(mc->text));
+					}
+				}
+			} else {
+				if (pipelining){
+					bgpq_pipeline(b, NULL, NULL, "!s%s\n",
+						b->defaultsources);
+					if (b->family == AF_INET)
+						bgpq_pipeline(b, bgpq_expanded_prefix,
+							NULL, "!i%s,1\n", bgpq_get_rset(mc->text));
+					else
+						bgpq_pipeline(b, bgpq_expanded_v6prefix,
+							NULL, "!i%s,1\n", bgpq_get_rset(mc->text));
+				} else {
+					bgpq_expand_irrd(b, NULL, NULL, "!s%s\n",
+						b->defaultsources);
+					if (b->family == AF_INET)
+						bgpq_expand_irrd(b, bgpq_expanded_prefix,
+							NULL, "!i%s,1\n", bgpq_get_rset(mc->text));
+					else
+						bgpq_expand_irrd(b, bgpq_expanded_v6prefix,
+							NULL, "!i%s,1\n", bgpq_get_rset(mc->text));
+				}
+			}
 		}
 
 		RB_FOREACH(asne, asn_tree, &b->asnlist) {
@@ -1093,14 +1376,16 @@ bgpq_expand(struct bgpq_expander *b)
 		fl &= ~O_NONBLOCK;
 		fcntl(fd, F_SETFL, fl);
 	}
+
 	close(fd);
+	free(b->defaultsources);
 
 	return 1;
 }
 
 void
-sx_radix_node_freeall(struct sx_radix_node *n) {
-
+sx_radix_node_freeall(struct sx_radix_node *n)
+{
 	if (n->l != NULL)
 		sx_radix_node_freeall(n->l);
 
@@ -1113,26 +1398,29 @@ sx_radix_node_freeall(struct sx_radix_node *n) {
 	if (n->payload)
 		free(n->payload);
 
-	sx_prefix_destroy(n->prefix);
+	sx_prefix_free(n->prefix);
 
 	free(n);
 }
 
 void
-sx_radix_tree_freeall(struct sx_radix_tree *t) {
-
+sx_radix_tree_freeall(struct sx_radix_tree *t)
+{
 	if (t->head != NULL)
 		sx_radix_node_freeall(t->head);
 
 	free(t);
 }
 
+/* XXX: needs cleaning up / figuring out */
 void
-bgpq_prequest_freeall(struct bgpq_prequest *bpr) {
+bgpq_prequest_freeall(struct bgpq_prequest *bpr)
+{
 }
 
 void
-expander_freeall(struct bgpq_expander *expander) {
+expander_freeall(struct bgpq_expander *expander)
+{
 	struct sx_tentry	*var, *nxt;
 	struct asn_entry	*asne, *asne_next;
 
