@@ -51,7 +51,7 @@ static int
 usage(int ecode)
 {
 	printf("\nUsage: bgpq4 [-h host[:port]] [-S sources] [-E|G|H <num>"
-	    "|f <num>|t] [-46ABbdJjKNnwXz] [-R len] <OBJECTS> ... "
+	    "|f <num>|t] [-46ABbdJjKNnpwXz] [-R len] <OBJECTS> ... "
 	    "[EXCEPT <OBJECTS> ...]\n");
 	printf("\nVendor targets:\n");
 	printf(" no option : Cisco IOS Classic (default)\n");
@@ -60,7 +60,8 @@ usage(int ecode)
 	printf(" -u        : Huawei XPL\n");
 	printf(" -j        : JSON\n");
 	printf(" -J        : Juniper Junos\n");
-	printf(" -K        : MikroTik RouterOS\n");
+	printf(" -K        : MikroTik RouterOSv6\n");
+	printf(" -K7       : MikroTik RouterOSv7\n");
 	printf(" -b        : NIC.CZ BIRD\n");
 	printf(" -N        : Nokia SR OS (Classic CLI)\n");
 	printf(" -n        : Nokia SR OS (MD-CLI)\n");
@@ -133,9 +134,9 @@ static void
 vendor_exclusive(void)
 {
 	fprintf(stderr, "-b (BIRD), -B (OpenBGPD), -F (formatted), -J (Junos),"
-	    " -j (JSON), -N (Nokia SR OS Classic), -n (Nokia SR OS MD-CLI),"
-	    " -U (Huawei), -u (Huawei XPL), -e (Arista) and -X (IOS XR) options "
-	    " are mutually exclusive\n");
+	    " -j (JSON), -K[7] (Microtik ROS), -N (Nokia SR OS Classic),"
+	    " -n (Nokia SR OS MD-CLI), -U (Huawei), -u (Huawei XPL),"
+	    "-e (Arista) and -X (IOS XR) options are mutually exclusive\n");
 	exit(1);
 }
 
@@ -190,14 +191,14 @@ main(int argc, char* argv[])
 		exit(1);
 	}
 #endif
-	
+
 	bgpq_expander_init(&expander, af);
 
 	if (getenv("IRRD_SOURCES"))
 		expander.sources=getenv("IRRD_SOURCES");
 
 	while ((c = getopt(argc, argv,
-	    "46a:AbBdDEeF:S:jJKf:l:L:m:M:NnW:pr:R:G:H:tTh:UuwXsvz")) != EOF) {
+	    "467a:AbBdDEeF:S:jJKf:l:L:m:M:NnpW:r:R:G:H:tTh:UuwXsvz")) != EOF) {
 	switch (c) {
 	case '4':
 		/* do nothing, expander already configured for IPv4 */
@@ -218,6 +219,13 @@ main(int argc, char* argv[])
 		expander.family = AF_INET6;
 		expander.tree->family = AF_INET6;
 		break;
+	case '7':
+		if (expander.vendor != V_MIKROTIK6) {
+			sx_report(SX_FATAL, "'7' can only be used after -K\n");
+			exit(1);
+		}
+		expander.vendor = V_MIKROTIK7;
+		break;
 	case 'a':
 		parseasnumber(&expander, optarg);
 		break;
@@ -297,10 +305,7 @@ main(int argc, char* argv[])
 	case 'K':
 		if (expander.vendor)
 			vendor_exclusive();
-		expander.vendor = V_MIKROTIK;
-		break;
-	case 'p':
-		expand_special_asn = 1;
+		expander.vendor = V_MIKROTIK6;
 		break;
 	case 'r':
 		refineLow = strtoul(optarg, NULL, 10);
@@ -389,6 +394,9 @@ main(int argc, char* argv[])
 			vendor_exclusive();
 		expander.vendor = V_NOKIA_MD;
 		break;
+	case 'p':
+		expand_special_asn = 1;
+		break;
 	case 't':
 		if (expander.generation)
 			exclusive();
@@ -451,7 +459,8 @@ main(int argc, char* argv[])
 			switch (vendor) {
 			case V_ARISTA:
 			case V_CISCO:
-			case V_MIKROTIK:
+			case V_MIKROTIK6:
+			case V_MIKROTIK7:
 				expander.aswidth = 4;
 				break;
 			case V_CISCO_XR:
@@ -694,17 +703,23 @@ main(int argc, char* argv[])
 		usage(1);
 
 	while (argv[0]) {
+		char *obj = argv[0];
+		char *delim = strstr(argv[0], "::");
+		if (delim) {
+			expander.usesource = 1;
+			obj = delim + 2;
+		}
 		if (!strcmp(argv[0], "EXCEPT")) {
 			exceptmode = 1;
 		} else if (exceptmode) {
 			bgpq_expander_add_stop(&expander, argv[0]);
-		} else if (!strncasecmp(argv[0], "AS-", 3)) {
+		} else if (!strncasecmp(obj, "AS-", 3)) {
 			bgpq_expander_add_asset(&expander, argv[0]);
-		} else if (!strncasecmp(argv[0], "RS-", 3)) {
+		} else if (!strncasecmp(obj, "RS-", 3)) {
 			bgpq_expander_add_rset(&expander, argv[0]);
-		} else if (!strncasecmp(argv[0], "AS", 2)) {
+		} else if (!strncasecmp(obj, "AS", 2)) {
 			char *ec;
-			if ((ec = strchr(argv[0], ':'))) {
+			if ((ec = strchr(obj, ':'))) {
 				if (!strncasecmp(ec + 1, "AS-", 3)) {
 					bgpq_expander_add_asset(&expander, argv[0]);
 				} else if (!strncasecmp(ec + 1, "RS-", 3)) {
